#![allow(dead_code)]

use crate::datastore::Store;
use challenge::NodeUpdate;
use std::pin::Pin;
use std::sync::Arc;
use std::thread::JoinHandle;
use tokio::sync::broadcast::Sender;
use tokio_stream::{Stream, StreamExt};
use tonic::{transport::Server, Request, Response, Status, Streaming};

pub mod challenge {
    tonic::include_proto!("challenge");
}

pub struct MyServer {
    ds: Arc<Store>,
    tx: Sender<NodeUpdate>,
}

impl MyServer {
    fn init(ds: Arc<Store>, tx: Sender<NodeUpdate>) -> Self {
        Self { ds, tx }
    }

    fn start(self) -> JoinHandle<()> {
        std::thread::spawn(|| {
            tokio::runtime::Runtime::new().unwrap().block_on(async {
                let addr = "0.0.0.0:31373".parse().unwrap();
                if let Err(e) = Server::builder()
                    .add_service(challenge::update_server::UpdateServer::new(self))
                    .serve(addr)
                    .await
                {
                    println!("gRPC server failed: {e:?}");
                };
            });
        })
    }
}

#[tonic::async_trait]
impl challenge::update_server::Update for MyServer {
    type GetUpdatesStream = Pin<Box<dyn Stream<Item = Result<NodeUpdate, Status>> + Send>>;

    async fn get_updates(
        &self,
        req: Request<Streaming<NodeUpdate>>,
    ) -> Result<Response<Self::GetUpdatesStream>, Status> {
        let remote_ip = req.remote_addr().unwrap().ip().to_string();
        let tx = self.tx.clone();
        let mut rx = self.tx.subscribe();
        let mut inbound = req.into_inner();

        let stream = async_stream::stream! {
            loop {
                tokio::select! {
                    Some(msg) = inbound.next() => {
                        match msg {
                            Ok(update) => {
                                if let Err(_) = tx.send(update.clone()) {
                                    yield Err(Status::internal("Failed to broadcast update"));
                                }
                            }
                            Err(e) => {
                                yield Err(Status::internal(format!("Error receiving client stream: {}", e)));
                                break;
                            }
                        }
                    }
                    Ok(update) = rx.recv() => {
                        if update.ip != remote_ip {
                            yield Ok(update);
                        }
                    }

                }
            }
        };

        Ok(Response::new(Box::pin(stream) as Self::GetUpdatesStream))
    }
}
